@@ -1,4 +1,3 @@
-from collections import ChainMap
 from dataclasses import dataclass, field
 from typing import Any, MutableMapping
 
@@ -8,7 +7,7 @@ class InstanceMap:
     _map: MutableMapping[Any, Any] = field(default_factory=dict)
 
     def chain(self) -> "InstanceMap":
-        return InstanceMap(ChainMap({}, self._map))
+        return InstanceMap({**self._map})
 
     def __contains__(self, target: Any) -> bool:
         return target in self._map
@@ -19,5 +18,8 @@ class InstanceMap:
     def __setitem__(self, target: Any, value: Any) -> None:
         self._map[target] = value
 
+    def __delitem__(self, target: Any) -> None:
+        del self._map[target]
+
     def clear(self) -> None:
         self._map.clear()
