@@ -8,10 +8,12 @@ from pydantic import BaseModel
 # Import of assert_type not available from typing until 3.11
 from typing_extensions import Annotated, assert_type
 
-from flexdi import CycleError, FlexGraph
+from flexdi import CycleError, FlexGraph, implicitbinding
+from flexdi.errors import ImplicitBindingError
 
 
-def test_dependant_simple() -> None:
+@pytest.mark.asyncio
+async def test_dependant_simple() -> None:
     """
     For a simple class with no arguments, we should be able to easily determine
     the dependency tree.
@@ -20,16 +22,20 @@ def test_dependant_simple() -> None:
     class Foo:
         pass
 
-    with FlexGraph() as graph:
-        res = graph.resolve(Foo)
+    async with FlexGraph() as graph:
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
 
 
-def test_dependant_chained() -> None:
+@pytest.mark.asyncio
+async def test_dependant_chained() -> None:
     """
     When arguments are specified, follow the tree to construct them.
     """
 
+    graph = FlexGraph()
+
+    @graph.bind
     class Foo:
         pass
 
@@ -37,17 +43,21 @@ def test_dependant_chained() -> None:
         def __init__(self, dep1: Foo) -> None:
             self.dep1 = dep1
 
-    with FlexGraph() as graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
 
 
-def test_dependant_chained_dataclasses() -> None:
+@pytest.mark.asyncio
+async def test_dependant_chained_dataclasses() -> None:
     """
     Now lets try it with dataclasses.
     """
 
+    graph = FlexGraph()
+
+    @graph.bind
     @dataclass
     class Foo:
         pass
@@ -56,39 +66,48 @@ def test_dependant_chained_dataclasses() -> None:
     class Bar:
         dep1: Foo
 
-    with FlexGraph() as graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
 
 
-def test_dependant_chained_pydantic() -> None:
+@pytest.mark.asyncio
+async def test_dependant_chained_pydantic() -> None:
     """
     Now lets try it with pydantic models used.
     """
 
+    graph = FlexGraph()
+
+    @graph.bind
     class Foo(BaseModel):
         pass
 
     class Bar(BaseModel):
         dep1: Foo
 
-    with FlexGraph() as graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
 
 
-def test_cycle_detection() -> None:
+@pytest.mark.asyncio
+async def test_cycle_detection() -> None:
     """
     This is a bit of a contrived case, but if there are cycles present in the
     dependency tree, we should fail as there would be no way to construct the
     object required.
     """
 
+    graph = FlexGraph()
+
+    @graph.bind
     class Foo:
         pass
 
+    @graph.bind
     class Bar:
         pass
 
@@ -101,14 +120,13 @@ def test_cycle_detection() -> None:
     Foo.__init__ = foo_init  # type: ignore
     Bar.__init__ = bar_init  # type: ignore
 
-    with FlexGraph() as graph:
-        with pytest.raises(CycleError):
-            graph.resolve(Foo)
-        with pytest.raises(CycleError):
-            graph.resolve(Bar)
+    with pytest.raises(CycleError):
+        async with graph:
+            pass
 
 
-def test_provider() -> None:
+@pytest.mark.asyncio
+async def test_provider() -> None:
     @dataclass
     class Foo:
         value: int
@@ -123,14 +141,15 @@ def test_provider() -> None:
     def foo_provider() -> Foo:
         return Foo(2)
 
-    with graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
         assert res.dep1.value == 2
 
 
-def test_provider_no_decorator() -> None:
+@pytest.mark.asyncio
+async def test_provider_no_decorator() -> None:
     @dataclass
     class Foo:
         value: int
@@ -145,14 +164,15 @@ def test_provider_no_decorator() -> None:
     graph = FlexGraph()
     graph.bind(foo_provider)
 
-    with graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
         assert res.dep1.value == 2
 
 
-def test_provider_lambda() -> None:
+@pytest.mark.asyncio
+async def test_provider_lambda() -> None:
     @dataclass
     class Foo:
         value: int
@@ -164,18 +184,23 @@ def test_provider_lambda() -> None:
     graph = FlexGraph()
     graph.bind(lambda: Foo(2), resolves=Foo)
 
-    with graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
         assert res.dep1.value == 2
 
 
-def test_provider_interface_map() -> None:
+@pytest.mark.asyncio
+async def test_provider_interface_map() -> None:
+    graph = FlexGraph()
+
+    @graph.bind
     @dataclass
     class Foo:
         pass
 
+    @graph.bind
     @dataclass
     class FooChild(Foo):
         pass
@@ -184,16 +209,19 @@ def test_provider_interface_map() -> None:
     class Bar:
         dep1: Foo
 
-    graph = FlexGraph()
     graph.bind(FooChild, resolves=Foo)
 
-    with graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, FooChild)
 
 
-def test_provider_lower_level() -> None:
+@pytest.mark.asyncio
+async def test_provider_lower_level() -> None:
+    graph = FlexGraph()
+
+    @graph.bind
     @dataclass
     class Foo:
         value: int
@@ -202,20 +230,19 @@ def test_provider_lower_level() -> None:
     class Bar:
         dep1: Foo
 
-    graph = FlexGraph()
-
     @graph.bind
     def value_provider() -> int:
         return 2
 
-    with graph:
-        res = graph.resolve(Bar)
+    async with graph:
+        res = await graph.resolve(Bar)
         assert isinstance(res, Bar)
         assert isinstance(res.dep1, Foo)
         assert res.dep1.value == 2
 
 
-def test_provider_clazz_mapping() -> None:
+@pytest.mark.asyncio
+async def test_provider_clazz_mapping() -> None:
     @dataclass
     class Foo:
         value: int
@@ -226,13 +253,14 @@ def test_provider_clazz_mapping() -> None:
     def value_provider() -> Any:
         return 3
 
-    with graph:
-        res = graph.resolve(Foo)
+    async with graph:
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value == 3
 
 
-def test_provider_annotated() -> None:
+@pytest.mark.asyncio
+async def test_provider_annotated() -> None:
     @dataclass
     class Foo:
         value1: Annotated[int, "value1"]
@@ -248,14 +276,15 @@ def test_provider_annotated() -> None:
     def value2_provider() -> Annotated[int, "value2"]:
         return 456
 
-    with graph:
-        res = graph.resolve(Foo)
+    async with graph:
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value1 == 123
         assert res.value2 == 456
 
 
-def test_provider_annotated_mapping() -> None:
+@pytest.mark.asyncio
+async def test_provider_annotated_mapping() -> None:
     @dataclass
     class Foo:
         value1: Annotated[int, "value1"]
@@ -271,14 +300,15 @@ def test_provider_annotated_mapping() -> None:
     def value2_provider() -> int:
         return 456
 
-    with graph:
-        res = graph.resolve(Foo)
+    async with graph:
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value1 == 123
         assert res.value2 == 456
 
 
-def test_provider_annotated_chained() -> None:
+@pytest.mark.asyncio
+async def test_provider_annotated_chained() -> None:
     @dataclass
     class Foo:
         value1: Annotated[int, "value1"]
@@ -294,14 +324,15 @@ def test_provider_annotated_chained() -> None:
     def value2_provider(value1: Annotated[int, "value1"]) -> int:
         return value1 + 111
 
-    with graph:
-        res = graph.resolve(Foo)
+    async with graph:
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value1 == 123
         assert res.value2 == 234
 
 
-def test_singleton() -> None:
+@pytest.mark.asyncio
+async def test_singleton() -> None:
     @dataclass
     class Foo:
         value1: int
@@ -324,26 +355,27 @@ def test_singleton() -> None:
         called_value2_provider += 1
         return "2"
 
-    with graph:
-        with graph.chain() as sub_injector:
-            sub_injector.resolve(Foo)
-        with graph.chain() as sub_injector:
-            sub_injector.resolve(Foo)
+    async with graph:
+        async with graph.chain() as sub_injector:
+            await sub_injector.resolve(Foo)
+        async with graph.chain() as sub_injector:
+            await sub_injector.resolve(Foo)
 
     assert called_value1_provider == 1
     assert called_value2_provider == 2
 
-    with graph:
-        with graph.chain() as sub_injector:
-            sub_injector.resolve(Foo)
-        with graph.chain() as sub_injector:
-            sub_injector.resolve(Foo)
+    async with graph:
+        async with graph.chain() as sub_injector:
+            await sub_injector.resolve(Foo)
+        async with graph.chain() as sub_injector:
+            await sub_injector.resolve(Foo)
 
     assert called_value1_provider == 2
     assert called_value2_provider == 4
 
 
-def test_sync_dep_provider() -> None:
+@pytest.mark.asyncio
+async def test_sync_dep_provider() -> None:
     @dataclass
     class Foo:
         value: int
@@ -357,10 +389,10 @@ def test_sync_dep_provider() -> None:
         provider_events.append("entered")
         return 1
 
-    with graph:
+    async with graph:
         provider_events = []
 
-        res = graph.resolve(Foo)
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value == 1
 
@@ -369,7 +401,8 @@ def test_sync_dep_provider() -> None:
     provider_events = ["entered"]
 
 
-def test_async_dep_provider() -> None:
+@pytest.mark.asyncio
+async def test_async_dep_provider() -> None:
     @dataclass
     class Foo:
         value: int
@@ -383,10 +416,10 @@ def test_async_dep_provider() -> None:
         provider_events.append("entered")
         return 1
 
-    with graph:
+    async with graph:
         provider_events = []
 
-        res = graph.resolve(Foo)
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value == 1
 
@@ -395,7 +428,8 @@ def test_async_dep_provider() -> None:
     provider_events = ["entered"]
 
 
-def test_sync_gen_provider() -> None:
+@pytest.mark.asyncio
+async def test_sync_gen_provider() -> None:
     @dataclass
     class Foo:
         value: int
@@ -412,10 +446,10 @@ def test_sync_gen_provider() -> None:
         finally:
             provider_events.append("exited")
 
-    with graph:
+    async with graph:
         provider_events = []
 
-        res = graph.resolve(Foo)
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value == 1
 
@@ -424,7 +458,8 @@ def test_sync_gen_provider() -> None:
     provider_events = ["entered", "exited"]
 
 
-def test_async_gen_provider() -> None:
+@pytest.mark.asyncio
+async def test_async_gen_provider() -> None:
     @dataclass
     class Foo:
         value: int
@@ -441,10 +476,10 @@ def test_async_gen_provider() -> None:
         finally:
             provider_events.append("exited")
 
-    with graph:
+    async with graph:
         provider_events = []
 
-        res = graph.resolve(Foo)
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.value == 1
 
@@ -453,43 +488,8 @@ def test_async_gen_provider() -> None:
     provider_events = ["entered", "exited"]
 
 
-def test_all_providers() -> None:
-    @dataclass
-    class Foo:
-        val1: Annotated[int, "value1"]
-        val2: Annotated[int, "value2"]
-        val3: Annotated[int, "value3"]
-        val4: Annotated[int, "value4"]
-
-    graph = FlexGraph()
-
-    @graph.bind(resolves=Annotated[int, "value1"])
-    def value1() -> int:
-        return 1
-
-    @graph.bind(resolves=Annotated[int, "value2"])
-    async def value2() -> int:
-        return 2
-
-    @graph.bind(resolves=Annotated[int, "value3"])
-    def value3() -> Iterator[int]:
-        yield 3
-
-    @graph.bind(resolves=Annotated[int, "value4"])
-    async def value4() -> AsyncIterator[int]:
-        yield 4
-
-    with graph:
-        res = graph.resolve(Foo)
-        assert isinstance(res, Foo)
-        assert res.val1 == 1
-        assert res.val2 == 2
-        assert res.val3 == 3
-        assert res.val4 == 4
-
-
 @pytest.mark.asyncio
-async def test_async_all_providers() -> None:
+async def test_all_providers() -> None:
     @dataclass
     class Foo:
         val1: Annotated[int, "value1"]
@@ -516,7 +516,7 @@ async def test_async_all_providers() -> None:
         yield 4
 
     async with graph:
-        res = await graph.aresolve(Foo)
+        res = await graph.resolve(Foo)
         assert isinstance(res, Foo)
         assert res.val1 == 1
         assert res.val2 == 2
@@ -524,7 +524,11 @@ async def test_async_all_providers() -> None:
         assert res.val4 == 4
 
 
-def test_supports_all_func_invocations_sync() -> None:
+@pytest.mark.asyncio
+async def test_supports_all_func_invocations() -> None:
+    graph = FlexGraph()
+
+    @graph.bind
     class Foo:
         pass
 
@@ -540,11 +544,11 @@ def test_supports_all_func_invocations_sync() -> None:
     async def func4(foo: Foo) -> AsyncIterator[Foo]:
         yield foo
 
-    with FlexGraph() as graph:
-        res1 = graph.resolve(func1)
-        res2 = graph.resolve(func2)
-        res3 = graph.resolve(func3)
-        res4 = graph.resolve(func4)
+    async with graph:
+        res1 = await graph.resolve(func1)
+        res2 = await graph.resolve(func2)
+        res3 = await graph.resolve(func3)
+        res4 = await graph.resolve(func4)
         assert_type(res1, Foo)
         assert_type(res2, Foo)
         assert_type(res3, Foo)
@@ -556,38 +560,7 @@ def test_supports_all_func_invocations_sync() -> None:
 
 
 @pytest.mark.asyncio
-async def test_supports_all_func_invocations_async() -> None:
-    class Foo:
-        pass
-
-    def func1(foo: Foo) -> Foo:
-        return foo
-
-    async def func2(foo: Foo) -> Foo:
-        return foo
-
-    def func3(foo: Foo) -> Iterator[Foo]:
-        yield foo
-
-    async def func4(foo: Foo) -> AsyncIterator[Foo]:
-        yield foo
-
-    async with FlexGraph() as graph:
-        res1 = await graph.aresolve(func1)
-        res2 = await graph.aresolve(func2)
-        res3 = await graph.aresolve(func3)
-        res4 = await graph.aresolve(func4)
-        assert_type(res1, Foo)
-        assert_type(res2, Foo)
-        assert_type(res3, Foo)
-        assert_type(res4, Foo)
-        assert isinstance(res1, Foo)
-        assert isinstance(res2, Foo)
-        assert isinstance(res3, Foo)
-        assert isinstance(res4, Foo)
-
-
-def test_async_gen_picks_correct_loop() -> None:
+async def test_gen_shutdown_order_loop() -> None:
     """
     Assert that when running asynchronous dependencies from a synchronous
     function that the cleanup is performed in the expected order.
@@ -619,13 +592,14 @@ def test_async_gen_picks_correct_loop() -> None:
     async def func3(bar: Bar) -> None:
         events.append(3)
 
-    with graph:
-        graph.resolve(func3)
+    async with graph:
+        await graph.resolve(func3)
 
     assert events == [1, 2, 3, 4, 5]
 
 
-def test_bindings_out_of_order() -> None:
+@pytest.mark.asyncio
+async def test_bindings_out_of_order() -> None:
     """
     If we get dependencies in a different order than their dependency resolution
     order, we should be able to construct the correct graph regardless.
@@ -648,6 +622,7 @@ def test_bindings_out_of_order() -> None:
     class Bar(Foo):
         pass
 
+    @graph.bind
     @dataclass
     class Fizz:
         foo: Foo
@@ -659,8 +634,241 @@ def test_bindings_out_of_order() -> None:
     graph.bind(Buzz)
     graph.bind(Bar, resolves=Foo)
 
-    with graph:
-        buzz = graph.resolve(Buzz)
+    async with graph:
+        buzz = await graph.resolve(Buzz)
         assert isinstance(buzz, Buzz)
         assert isinstance(buzz.fizz, Fizz)
         assert isinstance(buzz.fizz.foo, Bar)
+
+
+def test_entrypoints() -> None:
+    graph = FlexGraph()
+
+    @graph.bind
+    class Foo:
+        pass
+
+    @graph.entrypoint
+    def main1(foo: Foo) -> Foo:
+        return foo
+
+    @graph.entrypoint
+    async def main2(foo: Foo) -> Foo:
+        return foo
+
+    @graph.entrypoint()
+    def main3(foo: Foo) -> Foo:
+        return foo
+
+    @graph.entrypoint()
+    async def main4(foo: Foo) -> Foo:
+        return foo
+
+    res1 = main1()
+    res2 = main2()
+    res3 = main3()
+    res4 = main4()
+    assert_type(res1, Foo)
+    assert_type(res2, Foo)
+    assert_type(res3, Foo)
+    assert_type(res4, Foo)
+    assert isinstance(res1, Foo)
+    assert isinstance(res2, Foo)
+    assert isinstance(res3, Foo)
+    assert isinstance(res4, Foo)
+
+
+@pytest.mark.asyncio
+async def test_one_instance_multiple_targets() -> None:
+    """
+    Foo and Bar are normal classes.
+    A class Fizz exists which can satisfy either Foo or Bar.
+
+    We should be able to find the definition of Fizz to both Foo and Bar,
+    and there should only be one instance created from doing so.
+    """
+
+    graph = FlexGraph()
+
+    @dataclass
+    class Foo:
+        pass
+
+    @dataclass
+    class Bar:
+        pass
+
+    @dataclass
+    class Fizz(Foo, Bar):
+        pass
+
+    @dataclass
+    class Buzz:
+        foo: Foo
+        bar: Bar
+
+    graph.bind(Fizz, resolves=Foo)
+    graph.bind(Fizz, resolves=Bar)
+
+    async with graph:
+        buzz = await graph.resolve(Buzz)
+        assert isinstance(buzz.foo, Fizz)
+        assert isinstance(buzz.bar, Fizz)
+        assert id(buzz.foo) == id(buzz.bar)
+
+
+@pytest.mark.asyncio
+async def test_disallow_implicit_bindings_enter() -> None:
+    """
+    Bindings which are created with unresolvable values should cause the
+    application to raise an error on startup.
+    """
+
+    graph = FlexGraph()
+
+    @dataclass
+    class Foo:
+        pass
+
+    @graph.bind
+    @dataclass
+    class Bar:
+        foo: Foo
+
+    with pytest.raises(ImplicitBindingError):
+        async with graph:
+            pass
+
+
+@pytest.mark.asyncio
+async def test_disallow_implicit_bindings_resolve() -> None:
+    """
+    Resolving callables which have unresolvable values should cause the
+    application to raise an error on resolving.
+    """
+
+    graph = FlexGraph()
+
+    @dataclass
+    class Foo:
+        pass
+
+    @dataclass
+    class Bar:
+        foo: Foo
+
+    async with graph:
+        with pytest.raises(ImplicitBindingError):
+            await graph.resolve(Bar)
+
+
+@pytest.mark.asyncio
+async def test_allow_unbound_callable_with_explicit_bindings() -> None:
+    """
+    Ensure that resolving callables is still okay... they can be considered
+    explicit because we are directly passing them to resolve. If there are
+    any dependencies which it requires that are unbound, this should still
+    be considered an error.
+    """
+
+    graph = FlexGraph()
+
+    @graph.bind
+    @dataclass
+    class Foo:
+        pass
+
+    @dataclass
+    class Bar:
+        foo: Foo
+
+    async with graph:
+        res = await graph.resolve(Bar)
+        assert isinstance(res, Bar)
+        assert isinstance(res.foo, Foo)
+
+
+@pytest.mark.asyncio
+async def test_marked_implicitbinding() -> None:
+    """
+    When defining classes before creating an instance of the graph we can
+    declare that they are allowed to be implicitly bound to the dependency
+    tree. This helps reduce boilerplate in the main further for user classes
+    that are fully resolvable.
+    """
+
+    @implicitbinding
+    @dataclass
+    class Foo:
+        pass
+
+    @dataclass
+    class Bar:
+        foo: Foo
+
+    graph = FlexGraph()
+    async with graph:
+        res = await graph.resolve(Bar)
+        assert isinstance(res, Bar)
+        assert isinstance(res.foo, Foo)
+
+
+@pytest.mark.asyncio
+async def test_marked_implicitbinding_still_invalid() -> None:
+    """
+    Even if a direct dependency of your class is marked as an implicit binding,
+    it's sub-dependencies must still be bound, or be themselves implicit.
+
+    This tests the case where the sub-dependency is not bound.
+    """
+
+    @dataclass
+    class Foo:
+        pass
+
+    @implicitbinding
+    @dataclass
+    class Bar:
+        foo: Foo
+
+    @dataclass
+    class Fizz:
+        bar: Bar
+
+    graph = FlexGraph()
+    async with graph:
+        with pytest.raises(ImplicitBindingError):
+            await graph.resolve(Fizz)
+
+
+@pytest.mark.asyncio
+async def test_marked_implicitbinding_with_explicit_deps() -> None:
+    """
+    Even if a direct dependency of your class is marked as an implicit binding,
+    it's sub-dependencies must still be bound, or be themselves implicit.
+
+    This tests the case where the sub-dependency is bound explicitly.
+    """
+
+    @dataclass
+    class Foo:
+        pass
+
+    @implicitbinding
+    @dataclass
+    class Bar:
+        foo: Foo
+
+    @dataclass
+    class Fizz:
+        bar: Bar
+
+    graph = FlexGraph()
+    # Explicitly bind Foo.
+    graph.bind(Foo)
+
+    async with graph:
+        res = await graph.resolve(Fizz)
+        assert isinstance(res, Fizz)
+        assert isinstance(res.bar, Bar)
+        assert isinstance(res.bar.foo, Foo)
